/*
 * statemachine.cpp
 *
 *  Created on: 21 Sep 2023
 *      Author: Dario
 */

#include <sta/tacos/statemachine.hpp>

#include <sta/debug/debug.hpp>


namespace sta
{
	namespace tacos
	{
		Statemachine::Statemachine()
			: TacosThread{"Statemachine", STA_TACOS_STATEMACHINE_PRIORITY},
			  currentState_{STA_TACOS_INITIAL_STATE},
			  failsafeState_{STA_TACOS_INITIAL_STATE},
			  lockoutTimer_{[](void *){}, nullptr},
			  failsafeTimer_{[](void *){}, nullptr},
			  transitionFunc_{[](uint16_t) -> uint16_t { return Statemachine::instance()->getCurrentState(); }},
			  timerFunc_{[](uint16_t, uint16_t, uint16_t) {}}
		{
			STA_ASSERT(STA_TACOS_INITIAL_STATE < STA_TACOS_NUM_STATES);
		}

		void Statemachine::init()
		{
			timerFunc_(0, 0, EventFlags::STARTUP);
		}

		void Statemachine::func()
		{
			uint16_t next = transitionFunc_(currentState_);

			STA_ASSERT(next < STA_TACOS_NUM_STATES);

			// Check if a state change is desired. Block if the lockoutTimer is still running.
			if (next != currentState_ && !lockoutTimer_.isRunning())
			{
				if (failsafeTimer_.isRunning())
				{
					failsafeTimer_.stop();
				}

				// Call user space code to set the timers again.
				timerFunc_(next, currentState_, EventFlags::NORMAL);

				// Update the state and trigger the global state changed event.
				currentState_ = next;
				Statemachine::stateChangeEvent.set(EventFlags::NORMAL);
				Statemachine::stateChangeEvent.clear(EventFlags::ALL);
			}

			osThreadYield();
		}

		uint16_t Statemachine::getCurrentState() const
		{
			return currentState_;
		}

		void Statemachine::setStateFunction(state_fn func)
		{
			STA_ASSERT(func != nullptr);

			transitionFunc_ = func;
		}

		void Statemachine::setTimerFunction(timer_fn func)
		{
			STA_ASSERT(func != nullptr);

			timerFunc_ = func;
		}

		void Statemachine::setFailsafeTimer(uint32_t millis, uint16_t nextState)
		{
			STA_ASSERT(nextState < STA_TACOS_NUM_STATES);
			STA_ASSERT(!failsafeTimer_.isRunning());

			failsafeTimer_.setCallback([nextState](void* arg) {
				Statemachine::instance()->forceStateTransition(nextState);
			}, &failsafeState_);

			failsafeTimer_.start(millis);
		}

		void Statemachine::setLockoutTimer(uint32_t millis)
		{
			STA_ASSERT(!lockoutTimer_.isRunning());

			lockoutTimer_.setCallback([](void *) {  }, nullptr);
			lockoutTimer_.start(millis);
		}

		void Statemachine::forceStateTransition(uint16_t state)
		{
			if (failsafeTimer_.isRunning())
			{
				failsafeTimer_.stop();
			}

			if (lockoutTimer_.isRunning())
			{
				lockoutTimer_.stop();
			}

			timerFunc_(state, currentState_, EventFlags::FORCED);
			currentState_ = state;

			Statemachine::stateChangeEvent.set(EventFlags::FORCED);
			Statemachine::stateChangeEvent.clear(EventFlags::ALL);
		}

		Statemachine* Statemachine::_instance = nullptr;

		RtosEvent Statemachine::stateChangeEvent;
	}  // namespace tacos
}  // namespace sta



